/*
    Copyright (c) 2014 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS485p, imported from libdSvDC.

    dS485p is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS485p is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_AVAHI

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <string.h>
#include <pthread.h>

#include "discovery.h"
#include "base/log.h"
#include "base/util.h"

static int g_running;
static pthread_t avahi_tid;
static bool avahiDs485Enable;
static struct ds485_avahi_callbacks g_cb;

static void ds485p_avahi_resolve_callback(
        AvahiServiceResolver *resolver,
        AVAHI_GCC_UNUSED AvahiIfIndex interface,
        AVAHI_GCC_UNUSED AvahiProtocol protocol,
        AvahiResolverEvent event,
        const char *name,
        const char *type,
        const char *domain,
        const char *host_name,
        const AvahiAddress *address,
        uint16_t port,
        AVAHI_GCC_UNUSED AvahiStringList *txt,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        AVAHI_GCC_UNUSED void* userdata)
{
    if (!resolver) {
        return;
    }

    /* callback is called whenever a service has been resolved successfully or
       timed out */

    switch (event) {
    case AVAHI_RESOLVER_FAILURE:
        report(LOG_ERR, "failed to resolve service \"%s\" of type \"%s\" "
               "in domain \"%s\": %s\n", name, type, domain,
               avahi_strerror(avahi_client_errno(
                                                 avahi_service_resolver_get_client(resolver))));
        break;
    case AVAHI_RESOLVER_FOUND: {
        char ip[AVAHI_ADDRESS_STR_MAX];
        char *dsuid = NULL;
        avahi_address_snprint(ip, sizeof(ip), address);
        report(LOG_INFO, "found %s at: %s:%u (%s)\n", name, host_name,
               port, ip);

        AvahiStringList *pair = avahi_string_list_find(txt, "dSUID");
        if (pair) {
            dsuid = strchr((const char *)avahi_string_list_get_text(pair),
                          '=');
            if (dsuid) {
                // skip '='
                dsuid++;
            }
        }

        if (!dsuid) {
            report(LOG_DEBUG, "ignoring vdSM %s at: %s:%u (%s) - missing "
                   "dSUID in announcement\n", name, host_name, port, ip);
            break;
        }

        size_t buflen = /* tcp:// */ 6 + strlen(host_name) + /* : */ 1 +
                        /* 65535 */ 5 + /* NULL */ 1;
        char *uri = malloc(buflen);
        if (!uri) {
            report(LOG_ERR, "could not allocate memory for bus uri "
                   "(%s:%u / %s)\n", host_name, port, name);
            return;
        }
        snprintf(uri, buflen, "tcp://%s:%u", host_name, port);

        dsuid_t dsStruct;
        if (scanf_dsuid(dsuid, &dsStruct) != 0) {
            report(LOG_ERR, "failed to convert string to dsuid\n");
            return;
        }

        g_cb.bus_discovered(uri, name, &dsStruct);
        free(uri);
        break;
    }
    default:
        break;
    }

    avahi_service_resolver_free(resolver);
}

static void ds485p_avahi_browse_callback(
        AvahiServiceBrowser *browser,
        AvahiIfIndex interface,
        AvahiProtocol protocol,
        AvahiBrowserEvent event,
        const char *name,
        const char *type,
        const char *domain,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    if (!browser) {
        report(LOG_ERR, "avahi: invalid avahi service browser handle");
        return;
    }

    if (!userdata) {
        report(LOG_ERR, "avahi: missing user data\n");
        return;
    }

    ds485p_avahi_handle_t *avahi = userdata;

    /* called when a new services becomes available or is removed from LAN */

    switch (event) {
    case AVAHI_BROWSER_FAILURE:

        report(LOG_ERR, "avahi browse callback %s\n",
               avahi_strerror(avahi_client_errno(
                                  avahi_service_browser_get_client(browser))));
        avahi_simple_poll_quit(avahi->simple_poll);
        avahi->initialized = false;
        return;

    case AVAHI_BROWSER_NEW:
        report(LOG_DEBUG, "browser: new service \"%s\" of type \"%s\" in "
               "domain \"%s\"\n", name, type, domain);

        /* We ignore the returned resolver object. In the callback
           function we free it. If the server is terminated before
           the callback function is called the server will free
           the resolver for us. */

        if (!(avahi_service_resolver_new(avahi->client,
                                         interface, protocol, name, type, domain,
                                         AVAHI_PROTO_INET, 0, ds485p_avahi_resolve_callback,
                                         avahi->client)))
            report(LOG_ERR, "failed to resolve service \"%s\": %s\n", name,
                   avahi_strerror(avahi_client_errno(avahi->client)));

        break;

    case AVAHI_BROWSER_REMOVE:
        report(LOG_DEBUG, "browser: remove service \"%s\" of type \"%s\" "
               "in domain '%s'\n", name, type, domain);
        g_cb.bus_left(name);
        break;

    case AVAHI_BROWSER_ALL_FOR_NOW:
    case AVAHI_BROWSER_CACHE_EXHAUSTED:
        report(LOG_DEBUG, "browser: %s\n",
               event == AVAHI_BROWSER_CACHE_EXHAUSTED ? "CACHE_EXHAUSTED" : "ALL_FOR_NOW");
    default:
        break;
    }
}

static void ds485p_avahi_client_callback(
        AvahiClient *client,
        AvahiClientState state,
        AVAHI_GCC_UNUSED void * userdata)
{
    if (!client) {
        report(LOG_ERR, "avahi: invalid client handle\n");
        return;
    }

    if (!userdata) {
        report(LOG_ERR, "avahi: missing user data\n");
        return;
    }

    ds485p_avahi_handle_t *avahi = userdata;

    if (state == AVAHI_CLIENT_FAILURE) {
        report(LOG_ERR, "avahi server conection failure: %s\n",
               avahi_strerror(avahi_client_errno(client)));
        avahi_simple_poll_quit(avahi->simple_poll);
    }
}

bool ds485p_avahi_init(ds485p_avahi_handle_t *avahi, int enableDs485d)
{
    avahi->simple_poll = NULL;
    avahi->client = NULL;
    avahi->browser_vdc = NULL;
    avahi->browser_ds485 = NULL;
    avahi->initialized = false;

    int error;

    avahi->simple_poll = avahi_simple_poll_new();
    if (!avahi->simple_poll) {
        report(LOG_ERR, "could not allocate avahi_poll object\n");
        return false;
    }

    avahi->client = avahi_client_new(avahi_simple_poll_get(avahi->simple_poll),
                                     0, ds485p_avahi_client_callback,
                                     avahi, &error);
    if (!avahi->client) {
        report(LOG_ERR, "could not create avahi client: %s\n",
               avahi_strerror(error));
        avahi_simple_poll_free(avahi->simple_poll);
        avahi->simple_poll = NULL;
        return false;
    }

    /* avahi browser for vdsm */
    avahi->browser_vdc = avahi_service_browser_new(avahi->client, AVAHI_IF_UNSPEC,
                                                   AVAHI_PROTO_UNSPEC,
                                                   "_ds-vdsm._tcp",
                                                   NULL, 0,
                                                   ds485p_avahi_browse_callback,
                                                   avahi);

    if (!avahi->browser_vdc) {
        report(LOG_ERR, "could not create avahi browser for vdc: %s\n",
               avahi_strerror(avahi_client_errno(avahi->client)));
        avahi_client_free(avahi->client);
        avahi->client = NULL;
        avahi_simple_poll_free(avahi->simple_poll);
        avahi->simple_poll = NULL;
        return false;
    }

    /* avahi browser for ds485d */
    if (enableDs485d) {
        avahi->browser_ds485 = avahi_service_browser_new(avahi->client, AVAHI_IF_UNSPEC,
                                                         AVAHI_PROTO_UNSPEC,
                                                         "_ds-485d._tcp",
                                                         NULL, 0,
                                                         ds485p_avahi_browse_callback,
                                                         avahi);

        if (!avahi->browser_ds485) {
            report(LOG_ERR, "could not create avahi browser for ds485d: %s\n",
                   avahi_strerror(avahi_client_errno(avahi->client)));
            avahi_client_free(avahi->client);
            avahi->client = NULL;
            avahi_simple_poll_free(avahi->simple_poll);
            avahi->simple_poll = NULL;
            return false;
        }
    }

    avahi->initialized = true;

    return true;
}

void ds485p_avahi_shutdown(ds485p_avahi_handle_t *avahi)
{
    if (avahi->browser_vdc) {
        avahi_service_browser_free(avahi->browser_vdc);
    }

    if (avahi->browser_ds485) {
        avahi_service_browser_free(avahi->browser_ds485);
    }

    if (avahi->client) {
        avahi_client_free(avahi->client);
    }

    if (avahi->simple_poll) {
        avahi_simple_poll_free(avahi->simple_poll);
    }
}

static void*
sysAvahiThread(void *arg __attribute__((unused)))
{
    ds485p_avahi_handle_t avahi_handle;

    ds485p_avahi_init(&avahi_handle, avahiDs485Enable);

    while (g_running) {
        if (avahi_handle.initialized) {
            avahi_simple_poll_loop(avahi_handle.simple_poll);
            ds485p_avahi_shutdown(&avahi_handle);
            ds485p_avahi_init(&avahi_handle, avahiDs485Enable);
        } else {
            report(LOG_ERR, "avahi daemon interrupted or stopped. Trying to reconnect.\n");

            /* If ds485p_avahi_init fails we will retry again and again.
             * To avoid a busy loop, we will wait a second.
             */
            sleep(1);

            ds485p_avahi_init(&avahi_handle, avahiDs485Enable);
        }
    }

    ds485p_avahi_shutdown(&avahi_handle);
    return NULL;
}

int start_discovery(int filter, struct ds485_avahi_callbacks *cb)
{
    int ret;
    g_running = true;

    avahiDs485Enable = (filter & DISCOVER_DS485D);

    if (cb == NULL) {
        return -EINVAL;
    }
    memcpy(&g_cb, cb, sizeof(*cb));

    ret = pthread_create(&avahi_tid, NULL, sysAvahiThread, NULL);
    if (ret != 0) {
        report(LOG_ERR, "Could not create avahi thread.\n");
        return ret < 0 ? ret : -ret; /* make sure it's negative */
    }
    pthread_detach(avahi_tid);
    return 0;
}

void stop_discovery()
{
    g_running = false;
}
#endif//HAVE_AVAHI
